from typing import Literal

from pydantic import AnyHttpUrl, AnyUrl

from .base import IgnoringBaseModel, IgnoringBaseSettings


class ProxyOptions(IgnoringBaseModel):
    proxy_ca_bundle: str | None = None
    """
    The path to a custom certificate bundle to use when establishing SSL/TLS connections with proxy (optional).
    """

    proxy_client_cert: str | None = None
    """
    The path to a certificate for proxy TLS client authentication  (optional).
    """

    proxy_use_forwarding_for_https: bool = False
    """
    For HTTPS proxies, forward your requests to HTTPS destinations with an absolute URI. We strongly recommend you only use this option with trusted or corporate proxies.
    """


class S3Options(IgnoringBaseModel):
    endpoint_url: AnyHttpUrl
    """
    The URL for the S3 service.
    """

    bucket: str
    """
    The name of the S3 bucket.
    """

    access_key: str | None = None
    """
    The access key for the S3 bucket.
    If undefined, it is read from the AWS_ACCESS_KEY_ID environment variable.
    """

    secret: str | None = None
    """
    The secret key for the S3 bucket.
    If undefined, it is read from the AWS_SECRET_ACCESS_KEY environment variable.
    """

    session_token: str | None = None
    """
    The session token for temporary credentials (optional).
    """

    region_name: str | None = None
    """
    The region name for the S3 bucket.
    """

    api_version: str | None = None
    """
    The S3 API version.
    """

    use_ssl: bool = True
    """
    Whether to use SSL for S3 operations.
    """

    proxy_url: AnyUrl | None = None
    """
    The proxy URL for S3 operations (optional).
    """

    proxy_config: ProxyOptions | None = None
    """Proxy configuration for S3 operations."""

    request_checksum_calculation: Literal["when_supported", "when_required"] | None = None
    """
    Whether to calculate checksums for S3 request payloads (optional).
    Valid values are:

    * ``when_supported`` -- When set, a checksum will be calculated for
      all request payloads of operations modeled with the ``httpChecksum``
      trait where ``requestChecksumRequired`` is ``true`` or a
      ``requestAlgorithmMember`` is modeled.

    * ``when_required`` -- When set, a checksum will only be calculated
      for request payloads of operations modeled with the ``httpChecksum``
      trait where ``requestChecksumRequired`` is ``true`` or where a
      ``requestAlgorithmMember`` is modeled and supplied.

    Defaults to None.
    """

    multipart_chunksize: int = 256 * 1024**2
    """
    The size of the chunks to use for multipart uploads in bytes.
    """


class S3ConfigModel(IgnoringBaseSettings):
    s3: S3Options



"""
Validation methods for BAM files

e.g. warn on headers present, which can contain personally identifying
information
"""

import logging
from collections.abc import Generator
from os import PathLike

import pysam

log = logging.getLogger(__name__)


def validate_bam(bam_path: str | PathLike) -> Generator[str]:
    """Check if a BAM file is valid for submission"""
    errors: list[str] = []

    # disable pysam's SQ header presence enforcement (check_sq)
    #  these are used to optionally sort alignments by a custom contig order,
    #  which is unnecessary in unmapped BAM files
    bam_file = pysam.AlignmentFile(str(bam_path), mode="rb", check_sq=False)

    header = bam_file.header.to_dict()
    # ignore the HD key as it is almost always present and extremely
    # unlikely to accidentally contain identifying information
    concerning_keys = header.keys() - {"HD"}
    if concerning_keys:
        log.warning(f"Detected a header in BAM file '{bam_path}', ensure it contains no private information!")

    secondary_warned = False
    hard_clipped_warned = False
    # need until_eof since BAM won't have an index
    for read in bam_file.fetch(until_eof=True):
        if not secondary_warned and read.is_secondary:
            log.warning(
                f"Detected a secondary alignment in BAM file '{bam_path}'. "
                "Please consider filtering them to save upload bandwidth and storage space."
            )
            secondary_warned = True

        if not hard_clipped_warned and not read.is_secondary:
            hard_clipped_count = read.get_cigar_stats()[0][5]
            if hard_clipped_count:
                log.warning(
                    f"Detected hard-clipped base(s) in a primary alignment in BAM file '{bam_path}'. "
                    "This is a loss of information from the raw reads."
                )
                hard_clipped_warned = True

    yield from errors



"""
Module: logging

This module provides functions for setting up logging configuration.
"""

from __future__ import annotations

import logging
from os import PathLike
from pathlib import Path

log = logging.getLogger(__name__)

LOGGING_FORMAT = "%(asctime)s [%(levelname)s] %(name)s: %(message)s"
LOGGING_DATEFMT = "%Y-%m-%d %I:%M %p"


def add_filelogger(file_path: str | PathLike, level: str = "INFO", logger_name: str | None = None) -> None:
    """
    Add file logging for the specified package.

    This function configures a file logger to capture log messages
    for the package specified by logger_name. If no file path
    is provided, a default log file will be created in the user's
    home directory.

    :param file_path: Optional; the path to the log file. If None,
                      a default path will be used.
    :param level: Optional; the logging level. Default is 'INFO'.
                  Must be a valid logging level name (e.g., 'DEBUG', 'INFO').
    :param logger_name: Optional; the name of the logger to add the file handler to.
                        Default is the root logger.
    """
    # passing None to getLogger gets the root logger
    logger = logging.getLogger(logger_name)

    file_path = Path(file_path)

    file_handler = logging.FileHandler(file_path)
    file_handler.setLevel(level.upper())
    file_handler.setFormatter(logging.Formatter(fmt=LOGGING_FORMAT, datefmt=LOGGING_DATEFMT))
    logger.addHandler(file_handler)
    log.info(
        "File logger added for %s at %s with level %s.",
        logger.name,
        file_path,
        level.upper(),
    )


def setup_cli_logging(log_file: str | None, log_level: str):
    # set the root log level since this is the CLI
    logging.getLogger().setLevel(log_level.upper())

    logging.basicConfig(level=log_level.upper(), format=LOGGING_FORMAT, datefmt=LOGGING_DATEFMT)

    if log_file:
        # add file handler to root logger
        add_filelogger(
            log_file,
            log_level.upper(),
        )

    log.debug("Logging setup complete.")



"""Constants for progress bars, JSON schema validation, and other settings."""

TQDM_BAR_FORMAT = "{desc} ▕{bar:50}▏ {n_fmt:>10}/{total_fmt:<10} ({rate_fmt:>12}, ETA: {remaining:>6}) {postfix}"
TQDM_DEFAULTS = {
    "bar_format": TQDM_BAR_FORMAT,
    "unit": "iB",
    "unit_scale": True,
    "miniters": 1,
    "smoothing": 0.00001,
    "colour": "cyan",
}
